/*
 * Tetrons Copyright (c) 2006, James Bailie.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote
 * products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <sys/param.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#define LEFT 0
#define RIGHT 1
#define DOWN 2

#define OK 0
#define BLOCKED 1
#define STOPPED 2

#define PLAY 0
#define PAUSE 1
#define QUIT 2
#define MENU 3
#define EXIT 4

#define ON 1
#define OFF 0

#define CLEAR 1
#define NOCLEAR 0

#define X(v) ( v >> 4 )
#define Y(v) ( v & 0x0f )

#ifdef RANDOM
#define rand() random()
#endif

Display *display;
Visual *visual;
Window window;
XFontStruct *font;

GC blackGC, whiteGC, dottedGC;
XEvent event;
XTextProperty text_prop;

unsigned long colors[ 7 ];

char *name = "Tetrons";
char *font_names[ 5 ] =
{
   "-*-helvetica-medium-r-normal-*-*-140-*-*-*-iso8859-1",
   "-*-lucida-medium-r-normal-*-*-140-*-*-*-iso8859-1",
   "-*-clean-medium-r-normal-*-*-140-*-*-*-iso8859-1",
   "-*-fixed-medium-r-normal-*-*-140-*-*-*-iso8859-1",
   NULL
};

char *small_font_names[ 5 ] =
{
   "-*-helvetica-medium-r-normal-*-*-100-*-*-*-iso8859-1",
   "-*-lucida-medium-r-normal-*-*-100-*-*-*-iso8859-1",
   "-*-clean-medium-r-normal-*-*-100-*-*-*-iso8859-1",
   "-*-fixed-medium-r-normal-*-*-100-*-*-*-iso8859-1",
   NULL
};

XRectangle rectangles[ 4 ];

short int monochrome, score, screen, current, x, y, row, column,
   game_state, tetron_state, small_font, highscore;

char *highscore_file = "%s/.tetrons_high_score";

int grid_width, grid_height, block_size, small_block_size;

GC *grid[ 24 ][ 10 ];

/*
 * tetron.orientation:  orientation of tetron: 0-3.
 *
 * tetron.coordinates:  x, y, coordinates for each of the 4
 *                      blocks in the tetron, foreach of the 4 possible
 *                      orientations, relative to the upper left corner of
 *                      a 4 x 4 grid. 2 coordinates x 4 blocks x 4
 *                      positions = 32 integers, but we pack each
 *                      coordinate into a nibble of an unsigned char, so
 *                      this element is only 16 items in length.  These
 *                      coordinates get added to a base coordinate to
 *                      produce the world grid coordinate for each block in
 *                      the current tetron.
 */

struct tetron {
   GC gc;
   signed char orientation;
   unsigned char coordinates[ 16 ];
};

/*
 * 1 struct tetron for each of the 7 tetron types.
 */

struct tetron tetrons[ 7 ];

/*
 * A struct to for the hold area and the 2 preview tetrons.
 */

struct pending {
   char orientation, type;
};

struct pending pending[ 4 ];

/*
 * Data with which to initialize tetrons[].coordinates:
 *
 * Each row represents a tetron type.  Each group of eight columns in each
 * row represents one of the four possible orientations of that tetron.
 * The numbers are pairs of x, y, coordinates of individual blocks of the
 * tetrons, relative to the upper left corner of a 4 x 4 grid.
 */

unsigned char coordinates[] =
{
   1, 0, 1, 1, 1, 2, 2, 1,   1, 0, 0, 1, 1, 1, 2, 1,   1, 0, 0, 1, 1, 1, 1, 2,   0, 1, 1, 1, 2, 1, 1, 2,
   0, 0, 0, 1, 0, 2, 1, 2,   0, 1, 1, 1, 2, 1, 2, 0,   0, 0, 1, 0, 1, 1, 1, 2,   0, 1, 1, 1, 2, 1, 0, 2,
   0, 2, 1, 2, 1, 1, 1, 0,   0, 1, 1, 1, 2, 1, 2, 2,   0, 0, 1, 0, 0, 1, 0, 2,   0, 0, 0, 1, 1, 1, 2, 1,
   0, 2, 1, 2, 1, 1, 2, 1,   0, 0, 0, 1, 1, 1, 1, 2,   0, 2, 1, 2, 1, 1, 2, 1,   0, 0, 0, 1, 1, 1, 1, 2,
   0, 1, 1, 1, 1, 2, 2, 2,   0, 1, 0, 2, 1, 1, 1, 0,   0, 1, 1, 1, 1, 2, 2, 2,   0, 1, 0, 2, 1, 1, 1, 0,
   0, 0, 1, 0, 0, 1, 1, 1,   0, 0, 1, 0, 0, 1, 1, 1,   0, 0, 1, 0, 0, 1, 1, 1,   0, 0, 1, 0, 0, 1, 1, 1,
   1, 0, 1, 1, 1, 2, 1, 3,   0, 2, 1, 2, 2, 2, 3, 2,   1, 0, 1, 1, 1, 2, 1, 3,   0, 2, 1, 2, 2, 2, 3, 2
};

void reset_tetron();

/*
 * Initializes the tetrons[] array from the coordinates array.  We pack
 * each coordinate into a nibble of an unsigned char.
 */

void initialize_tetrons()
{
   int i, j;
   char *ptr, buffer[ MAXPATHLEN + 1 ];

   for( i = 0; i < 7; ++i )
      for( j = 0, ptr = &coordinates[ i * 32 ]; j < 16; ++j, ptr += 2 )
         tetrons[ i ].coordinates[ j ] = ( *ptr << 4 ) + *( ptr + 1 );

   pending[ 0 ].type = pending[ 1 ].type =
      pending[ 2 ].type = pending[ 3 ].type = -1;

   tetron_state = STOPPED;
#ifdef RANDOM
   srandomdev();
#else
   srand( time( NULL ) / ( rand() * 100000 ));
#endif

   highscore = 0;

   if (( ptr = getenv( "HOME" )) != NULL )
   {
      snprintf( buffer, sizeof( buffer ), highscore_file, ptr );

      if (( i = open( buffer, O_RDONLY )) >= 0 )
      {
         if (( j = read( i, buffer, sizeof( buffer ) - 1 )) > 0 )
         {
            buffer[ j ] = '\0';
            highscore = atoi( buffer );
         }
              
         close( i );
      }
   }
}

/*
 * Frees resources allocated in the X server, closes the connection,
 * saves the highscore.
 */

void cleanup_display()
{
   int i;
   mode_t mode;
   char *ptr, buffer[ MAXPATHLEN + 1 ];

   if ( font != NULL )
      XFreeFont( display, font );

   if ( ! monochrome )
   {
      XFreeColors( display, DefaultColormap( display, screen ), colors, 7, 0 );

      for( i = 0; i < 7; ++i )
         XFreeGC( display, tetrons[ i ].gc );
   }

   if ( blackGC )
      XFreeGC( display, blackGC );
   if ( whiteGC )
      XFreeGC( display, whiteGC );
   if ( dottedGC )
      XFreeGC( display, dottedGC );

   XDestroyWindow( display, window );
   XCloseDisplay( display );

   if (( ptr = getenv( "HOME" )) != NULL )
   {
      snprintf( buffer, sizeof( buffer ), highscore_file, ptr );
      mode = getmode( setmode( "0600" ), 0 );

      if (( i = open( buffer, O_WRONLY | O_CREAT | O_TRUNC, mode )) >= 0 )
      {
         snprintf( buffer, sizeof( buffer ), "%d", highscore );
         write( i, buffer, strlen( buffer ));
         close( i );
      }
   }
}

/*
 * Creates the game window.  Sets the window title.  Maps the window.
 */

void initialize_window()
{
   XSetWindowAttributes attributes;
   unsigned long attribute_mask;

   attributes.background_pixel = BlackPixel( display, screen );
   attributes.border_pixel = WhitePixel( display, screen );
   attributes.event_mask = KeyPressMask | ExposureMask | StructureNotifyMask;
   attribute_mask = CWBackPixel | CWBorderPixel | CWEventMask;

   window = XCreateWindow( display, RootWindow( display,screen ),
                           0,
                           0,
                           grid_width, grid_height,
                           0,
                           CopyFromParent,
                           InputOutput,
                           visual,
                           attribute_mask,
                           &attributes );

   if ( ! window )
   {
      fprintf( stderr, "Could not create window\n" );
      exit( 1 );
   }

   if ( XStringListToTextProperty( &name, 1, &text_prop ) )
      XSetWMName( display, window, &text_prop );

   XMapWindow( display, window );
   XFlush( display );

   for( ; ; )
   {
      XNextEvent( display, &event );

      if ( XFilterEvent( &event, window ) == True )
         continue;

      if ( event.type == MapNotify )
         break;
   }
}

/*
 * Sends the minimal window manager hints, and tells the WM we
 * would like to participate in the WM_DELETE_WINDOW protocol
 * so we can exit gracefully, if the WM wants us to (eg., the
 * user clicked a title bar close button).
 */

void send_wmhints()
{
   XWMHints wmhints;
   XSizeHints shints;
   Atom protocols, delete;

   wmhints.input = True;
   wmhints.initial_state = NormalState;
   wmhints.flags = InputHint | StateHint;

   XSetWMHints( display, window, &wmhints );

   shints.min_width = grid_width;
   shints.max_width = grid_width;
   shints.base_width = grid_width;

   shints.min_height = grid_height;
   shints.max_height = grid_height;
   shints.base_height = grid_height;

   shints.flags = PSize | PMinSize | PMaxSize;

   XSetWMNormalHints( display, window, &shints );

   protocols = XInternAtom( display, "WM_PROTOCOLS", False );
   delete = XInternAtom( display, "WM_DELETE_WINDOW", False );

   if ( protocols != None && delete != None )
      XSetWMProtocols( display, window, &delete, 1 );
}

/*
 * Creates graphic contexts we use.
 */

void initialize_gc()
{
   XGCValues gcvalues;
   unsigned long gcmask;
   struct tetron *tptr;
   char **ptr, *names[] = { "blue2", "red2", "yellow2", "green2", "purple2", "royalblue", "hotpink2" };
   int i;
   XColor color;

   blackGC = whiteGC = dottedGC = 0;

   gcvalues.function = GXcopy;
   gcvalues.line_width = 1;
   gcvalues.line_style = LineSolid;
   gcvalues.cap_style = CapButt;
   gcvalues.fill_style = FillSolid;
   gcvalues.fill_rule = EvenOddRule;
   gcvalues.background = BlackPixel( display, screen );

   gcmask = GCFunction | GCBackground | GCForeground | GCLineWidth | GCLineStyle | GCFillStyle | GCFillRule;

   for( i = 0, ptr = names, tptr = tetrons; i < 7; ++i, ++ptr, ++tptr )
   {
      if ( ! XAllocNamedColor( display, DefaultColormap( display, screen ), *ptr, &color, &color ))
      {
         monochrome = 1;
         XFreeColors( display, DefaultColormap( display, screen ), colors, i, 0 );
         break;
      }

      colors[ i ] = gcvalues.foreground = color.pixel;

      if ( ! ( tptr->gc = XCreateGC( display, window, gcmask, &gcvalues )))
      {
         XFreeColors( display, DefaultColormap( display, screen ), colors, i, 0 );
         monochrome = 1;
         break;
      }
   }

   if ( monochrome )
      while( --i >= 0 )
         XFreeGC( display, tetrons[ i ].gc );

   gcvalues.foreground = WhitePixel( display, screen );

   if ( ! ( whiteGC = XCreateGC( display, window, gcmask, &gcvalues )))
   {
      fprintf( stderr, "Could not create whiteGC" );
      cleanup_display();
      exit( 1 );
   }

   gcvalues.line_style = LineOnOffDash;

   if ( ! ( dottedGC = XCreateGC( display, window, gcmask, &gcvalues )))
   {
      fprintf( stderr, "Could not create dottedGC" );
      cleanup_display();
      exit( 1 );
   }

   gcvalues.line_style = LineSolid;
   gcvalues.foreground = BlackPixel( display, screen );
   gcvalues.background = WhitePixel( display, screen );

   if ( ! ( blackGC = XCreateGC( display, window, gcmask, &gcvalues )))
   {
      fprintf( stderr, "Could not create blackGC" );
      cleanup_display();
      exit( 1 );
   }
}

/*
 * Determines the size of the game window and the size of the blocks,
 * according to the resolution of the display.
 */

void initialize_screen()
{
   int width, height;

   width = DisplayWidth( display, screen );
   height = DisplayHeight( display, screen );

   switch ( height )
   {
      case 480:
         small_font = 1;
         grid_height = 461;
         grid_width = 323;
         small_block_size = 12;
         block_size = 23;
         break;

      case 600:
      case 768:
         small_font = 0;
         grid_height = 581;
         grid_width = 407;
         block_size = 29;
         small_block_size = 15;
         break;

      case 1024:
         small_font = 0;
         grid_height = 901;
         grid_width = 631;
         block_size = 45;
         small_block_size = 23;
         break;

      default:
         fprintf( stderr, "Unrecognized display resolution\n" );
         exit( 1 );
   }
}

/*
 * Initializes a single roman font for use in titles and menus.
 */

void initialize_font()
{
   char **ptr;

   for( ptr = ( small_font ? small_font_names : font_names ); *ptr; ++ptr )
      if (( font = XLoadQueryFont( display, *ptr ) ) != NULL )
         break;

   if ( font == NULL )
   {
      cleanup_display();
      fprintf( stderr, "Could not find a font to use\n" );
      exit( 1 );
   }

   XSetFont( display, whiteGC, font->fid );
}

/*
 * draws the main menu in the grid area.
 */

void draw_menu()
{
   char *keys[] = { "[ P ]", "[ Q ]", "[ UP ]", "[ DOWN ]", "[ LEFT ]", "[ RIGHT ]", "[ SPACE ]", "" };
   char *funcs[] = { "Pause / Play", "Quit", "Rotate Tetron", "Drop Tetron",
                     "Move Tetron Left", "Move Tetron Right", "Hold Tetron" };

   char *title = "TETRONS";
   char *author = "by jimmy@mammothcheese.ca";

   char *title2 ="Your Personal High Score";
   char **ptr, **ptr2;
   char buffer[ 128 ];

   int h, w, xbase, ybase, xwidth, len;

   h = font->ascent + font->descent + small_block_size;
   ybase = ( 20 * block_size - h * 13 ) / 2;

   w = XTextWidth( font, title, len = strlen( title ));

   XDrawString( display, window, whiteGC,
                ( 10 * block_size - w ) / 2,
                ybase, title, len );

   w = XTextWidth( font, author, len = strlen( author ));

   XDrawString( display, window, whiteGC,
                ( 10 * block_size - w ) / 2,
                ybase += h, author, len );

   w = XTextWidth( font, title2, len = strlen( title2 ));

   XDrawString( display, window, whiteGC,
                ( 10 * block_size - w ) / 2,
                ybase += h * 2, title2, len );

   snprintf( buffer, sizeof( buffer ), "%d", highscore );
   w = XTextWidth( font, title2, len = strlen( buffer ));

   XDrawString( display, window, whiteGC,
                ( 10 * block_size - w ) / 2,
                ybase += h, buffer, len );

   ybase += h * 2;

   len = XTextWidth( font, keys[ 6 ], strlen( keys[ 6 ] ));
   w = len + XTextWidth( font, funcs[ 5 ], strlen( funcs[ 5 ] )) + block_size;

   xbase = ( 10 * block_size - w ) / 2;
   xwidth = xbase + len + block_size;

   for( ptr = keys, ptr2 = funcs; **ptr; ++ptr, ++ptr2, ybase += h )
   {
      XDrawString( display, window, whiteGC,
                   xbase, ybase, *ptr, strlen( *ptr ));
      XDrawString( display, window, whiteGC,
                   xwidth, ybase, *ptr2, strlen( *ptr2 ));
   }

   XFlush( display );
}

/*
 * Adjusts the screen x coordinate for tetrons drawn in the side panel,
 * depending upon their orientation, so that they are centered in the
 * panel.
 */

int adjust_pending( int i )
{
   switch( pending[ i ].type )
   {
      case 0:
         if ( pending[ i ].orientation % 2 )
            return small_block_size / 2;
         else if ( pending[ i ].orientation == 0 )
            return 0;

      case 1:
      case 2:
         if ( pending[ i ].orientation % 2 )
            return small_block_size / 2;
         break;

      case 3:
      case 4:
         if ( pending[ i ].orientation % 2 == 0 )
            return small_block_size / 2;
         break;

      case 6:
         if ( pending[ i ].orientation % 2 )
            return 0;
         else return small_block_size / 2;
   }

   return small_block_size;
}

/*
 * Draws the pending tetrons in the side panel.
 */

void draw_pending()
{
   int i, p, lx, ly;
   struct tetron *t;

   lx = 11 * block_size;
   ly = 2 * block_size;

   XFillRectangle( display, window, blackGC,
                   lx, ly, lx + 4 * small_block_size, 18 * small_block_size );

   for( p = 0; p < 3; ++p )
   {
      t = &tetrons[ ( int )pending[ p ].type ];

      for( i = 0; i < 4; ++i )
      {
         rectangles[ i ].x = lx + X( t->coordinates[ pending[ p ].orientation * 4 + i ] ) * small_block_size + 2;
         rectangles[ i ].y = ly + Y( t->coordinates[ pending[ p ].orientation * 4 + i ] ) * small_block_size + 2;
         rectangles[ i ].width = small_block_size - 4;
         rectangles[ i ].height = small_block_size - 4;
         rectangles[ i ].x += adjust_pending( p );
      }

      XDrawRectangles( display, window, whiteGC, rectangles, 4 );

      if ( ! monochrome )
      {
         for( i = 0; i < 4; ++i )
         {
            rectangles[ i ].x = lx + X( t->coordinates[ pending[ p ].orientation * 4 + i ] ) * small_block_size + 3;
            rectangles[ i ].y = ly + Y( t->coordinates[ pending[ p ].orientation * 4 + i ] ) * small_block_size + 3;
            rectangles[ i ].width = small_block_size - 5;
            rectangles[ i ].height = small_block_size - 5;
            rectangles[ i ].x += adjust_pending( p );
         }

         XFillRectangles( display, window, t->gc, rectangles, 4 );
      }

      ly += 6 * small_block_size;
   }
}

/*
 * Draws the tetron in the hold area, in the side panel.
 */

void draw_held( int draw )
{
   int i, lx, ly;
   struct tetron *t;

   lx = 11 * block_size;
   ly = 25 * small_block_size;

   t = &tetrons[ ( int )pending[ 3 ].type ];

   for( i = 0; i < 4; ++i )
   {
      rectangles[ i ].x = lx + X( t->coordinates[ pending[ 3 ].orientation * 4 + i ] ) * small_block_size + 2;
      rectangles[ i ].y = ly + Y( t->coordinates[ pending[ 3 ].orientation * 4 + i ] ) * small_block_size + 2;
      rectangles[ i ].width = small_block_size - 4;
      rectangles[ i ].height = small_block_size - 4;
      rectangles[ i ].x += adjust_pending( 3 );
   }

   XDrawRectangles( display, window, ( draw ? whiteGC : blackGC ), rectangles, 4 );

   if ( ! monochrome )
   {
      if ( ! draw )
         XFillRectangles( display, window, blackGC, rectangles, 4 );
      else
      {
         for( i = 0; i < 4; ++i )
         {
            rectangles[ i ].x = lx + X( t->coordinates[ pending[ 3 ].orientation * 4 + i ] ) * small_block_size + 3;
            rectangles[ i ].y = ly + Y( t->coordinates[ pending[ 3 ].orientation * 4 + i ] ) * small_block_size + 3;
            rectangles[ i ].width = small_block_size - 5;
            rectangles[ i ].height = small_block_size - 5;
            rectangles[ i ].x += adjust_pending( 3 );
         }

         XFillRectangles( display, window, t->gc, rectangles, 4 );
      }
   }
}

/*
 * draws the current score in the side panel.
 */

void draw_score()
{
   static int w = 0, h = 0, len = 0;
   static char buffer[ 64 ];

   if ( len )
   {
      XFillRectangle( display, window, blackGC,
                      block_size * 10 + ( 4 * block_size - w ) / 2,
                      ( small_block_size * 31 + ( 9 * small_block_size - h ) / 2 ) - h,
                      w, h );
   }

   snprintf( buffer, sizeof( buffer ), "%d", score );

   h = font->ascent + font->descent;
   w = XTextWidth( font, buffer, len = strlen( buffer ));

   XDrawString( display, window, whiteGC,
                block_size * 10 + ( 4 * block_size - w ) / 2,
                small_block_size * 31 + ( 9 * small_block_size - h ) / 2,
                buffer, len );
}

/*
 * Draws the side panel.  Called after every time a new tetron
 * is generated to update the pending tetron list.
 */

void draw_game()
{
   int w;
   char *title1 = "NEXT", *title2 = "HELD", *title3 = "SCORE";

   w = 10 * block_size + 1;
   XDrawLine( display, window, dottedGC,
              w, 0, w, 20 * block_size );

   w = XTextWidth( font, "Next", 4 );

   XDrawString( display, window, whiteGC,
                block_size * 10 + ( 4 * block_size - w ) / 2,
                block_size, title1, 4 );

   if ( pending[ 0 ].type >= 0 )
      draw_pending();

   w = XTextWidth( font, "Held", 4 );

   XDrawString( display, window, whiteGC,
                block_size * 10 + ( 4 * block_size - w ) / 2,
                small_block_size * 23, title2, 4 );

   if ( pending[ 3 ].type >= 0 )
      draw_held( ON );

   w = XTextWidth( font, "Score", 5 );

   XDrawString( display, window, whiteGC,
                block_size * 10 + ( 4 * block_size - w ) / 2,
                small_block_size * 31, title3, 5 );

   draw_score();

   XFlush( display );
}

/*
 * Opens the connection to the X server, calls the display initialization
 * functions.  Recognizes the -display command-line argument.
 */

void initialize_display( int argc, char **argv )
{
   char **ptr, *disp;

   disp = NULL;

   for( ptr = argv; *ptr; ++ptr )
      if ( ! strcmp( *ptr, "-disp" ) && *( ptr + 1 ))
         disp = *++ptr;

   if (( display = XOpenDisplay( disp )) == NULL )
   {
      fprintf( stderr, "Could not open display\n" );
      exit( 1 );
   }

   screen = DefaultScreen( display );
   visual = CopyFromParent;

   initialize_screen();
   initialize_window();
   send_wmhints();

   monochrome = 0;
   initialize_gc();
   initialize_font();
}

/*
 * draw_tetron( ON ) draws the current tetron in white outline.
 * draw_tetron( OFF ) draws the current tetron in black outline,
 * erasing it.
 */

void draw_tetron( int draw )
{
   int i;

   for( i = 0; i < 4; ++i )
   {
      rectangles[ i ].x = x + X( tetrons[ current ].coordinates[ tetrons[ current ].orientation * 4 + i ] ) * block_size + 4;
      rectangles[ i ].y = y + Y( tetrons[ current ].coordinates[ tetrons[ current ].orientation * 4 + i ] ) * block_size + 4;
      rectangles[ i ].width = block_size - 8;
      rectangles[ i ].height = block_size - 8;
   }

   XDrawRectangles( display, window, ( draw ? whiteGC : blackGC ), rectangles, 4 );

   if ( ! monochrome )
   {
      if ( ! draw )
         XFillRectangles( display, window, blackGC, rectangles, 4 );
      else
      {
         for( i = 0; i < 4; ++i )
         {
            rectangles[ i ].x = x + X( tetrons[ current ].coordinates[ tetrons[ current ].orientation * 4 + i ] ) * block_size + 5;
            rectangles[ i ].y = y + Y( tetrons[ current ].coordinates[ tetrons[ current ].orientation * 4 + i ] ) * block_size + 5;
            rectangles[ i ].width = block_size - 9;
            rectangles[ i ].height = block_size - 9;
         }

         XFillRectangles( display, window, tetrons[ current ].gc, rectangles, 4 );
      }
   }
}

/*
 * The purpose of the following two functions is explained in draw_window().
 */

void draw_off_on()
{
   int i, j;

   for( i = 4; i < 24; ++i )
      for( j = 0; j < 10; ++j )
         if ( grid[ i ][ j ] == NULL )
            XFillRectangle( display, window, blackGC,
                            j * block_size, ( i - 4 ) * block_size,
                            block_size, block_size );
         else
         {
            XDrawRectangle( display, window, whiteGC,
                            j * block_size + 4, ( i - 4 ) * block_size + 4,
                            block_size - 8, block_size - 8 );

            if ( ! monochrome )
               XFillRectangle( display, window, *grid[ i ][ j ],
                               j * block_size + 5, ( i - 4 ) * block_size + 5,
                               block_size - 9, block_size - 9 );
         }
}

void draw_on()
{
   int i, j;

   for( i = 4; i < 24; ++i )
      for( j = 0; j < 10; ++j )
         if ( grid[ i ][ j ] != NULL )
         {
            XDrawRectangle( display, window, whiteGC,
                            j * block_size + 4, ( i - 4 ) * block_size + 4,
                            block_size - 8, block_size - 8 );

            if ( ! monochrome )
               XFillRectangle( display, window, *grid[ i ][ j ],
                               j * block_size + 5, ( i - 4 ) * block_size + 5,
                               block_size - 9, block_size - 9 );
         }
}

/*
 * draw_window( NOCLEAR ) redraws the world grid.
 * draw_window( CLEAR ) clears the entire game toplevel window,
 * then redraws the world grid.
 *
 * The first 4 rows of the world grid are offscreen.
 *
 * This function does not redraw the current tetron, as it does not become
 * part of the world grid until it is blocked.
 *
 * Nor does it redraw the side panel.  draw_game() does that.
 */

void draw_window( int clear )
{
   if ( clear )
      XClearWindow( display, window );

   if ( game_state == MENU )
      draw_menu();

   /*
    * If we have cleared the entire window, then we only need to draw the "on"
    * blocks.
    */

   else if ( clear )
      draw_on();
   else
      draw_off_on();
}

/*
 * check_motion( [ DOWN | LEFT | RIGHT ] ) returns OK if desired motion is
 * possible for current tetron.  Returns BLOCKED if the motion is not
 * possible, or STOPPED if the tetron is blocked while at least one of
 * its constituent blocks is not onscreen.
 */

int check_motion( int where )
{
   char *c, *limit;
   struct tetron *t;

   t = &tetrons[ current ];
   limit = &t->coordinates[ ( t->orientation + 1 ) * 4 ];

   for( c = &t->coordinates[ t->orientation * 4 ];
        c < limit;
        ++c )
      switch( where )
      {
         case LEFT:
            if ( grid[ Y( *c ) + row ][ X( *c ) + column - 1 ] != NULL ||
                 X( *c ) + column == 0 )
               return ( row < 4 ? STOPPED : BLOCKED );
            break;

         case RIGHT:
            if ( grid[ Y( *c ) + row ][ X( *c ) + column + 1 ] != NULL ||
                 X( *c ) + column == 9 )
               return ( row < 4 ? STOPPED : BLOCKED );
            break;

         case DOWN:
            if ( grid[ Y( *c ) + row + 1 ][ X( *c ) + column ] != NULL ||
                Y( *c ) + row == 23 )
               return ( row < 4 ? STOPPED : BLOCKED );
            break;
      }

   return OK;
}

/*
 * move_tetron( [ LEFT | RIGHT | DOWN ] ) moves the current tetron one
 * position on the grid, left, right or down.
 */

void move_tetron( int where )
{
   draw_tetron( OFF );

   switch( where )
   {
      case LEFT:
         x -= block_size;
         --column;
         break;

      case RIGHT:
         x += block_size;
         ++column;
         break;

      case DOWN:
         y += block_size;
         ++row;
         break;
   }

   draw_tetron( ON );
   XFlush( display );
}

/*
 * Adds the current tetron to the world grid.  Called after
 * increment_tetron() returns BLOCKED, indicating it can fall no farther.
 */

void add_tetron()
{
   struct tetron *t;
   char *c, *limit;

   t = &tetrons[ current ];
   limit = &t->coordinates[ ( t->orientation + 1 ) * 4 ];

   for( c = &t->coordinates[ t->orientation * 4 ];
        c < limit;
        ++c )
      grid[ row + Y( *c ) ][ column + X( *c  ) ] = &t->gc;
}

/*
 * Flashes a box around a complete line of blocks three times.
 */

void flash_complete( int i )
{
   int j, k;

   for( j = 0; j < 3; ++j )
   {
      XFillRectangle( display, window, blackGC,
                      0, ( i - 4 ) * block_size,
                      block_size * 10, block_size );

      XFlush( display );

      usleep( 100000 );

      for( k = 0; k < 10; ++k )
         if ( grid[ i ][ k ] != NULL )
         {
            XDrawRectangle( display, window, whiteGC,
                            k * block_size + 4, ( i - 4 ) * block_size + 4,
                            block_size - 8, block_size - 8 );
      
            if ( ! monochrome )
               XFillRectangle( display, window, *grid[ i ][ k ],
                               k * block_size + 5, ( i - 4 ) * block_size + 5,
                               block_size - 9, block_size - 9 );
         }
   
      XFlush( display );

      usleep( 100000 );
   }
}

/*
 * Removes completed lines of blocks from the grid.
 */

void collapse_complete()
{
   int i, j, i2, count;

   for( i = 23; i > 3; --i )
   {
      count = 0;

      for( j = 0; j < 10; ++j )
         if ( grid[ i ][ j ] != NULL )
            ++count;

      if ( count == 10 )
      {
         score += count;

         flash_complete( i );

         for( i2 = i; i2 > 0; --i2 )
            for( j = 0; j < 10; ++j )
               grid[ i2 ][ j ] = grid[ i2 - 1 ][ j ];

         draw_window( NOCLEAR );
         draw_score();
         XFlush( display );

         ++i;
      }
   }
}

/*
 * Calling increment_tetron() repeatedly, until it does not return OK,
 * as a means of dropping a tetron, causes screen flashes because the
 * tetron gets continually erased and redrawn, so we have a separate
 * function to drop the tetron to the bottom and update the screen once.
 */

void drop_tetron()
{
   draw_tetron( OFF );

   while(( tetron_state = check_motion( DOWN )) == OK )
   {
      y += block_size;
      ++row;
   }

   add_tetron();
   draw_tetron( ON );
   collapse_complete();
}

/*
 * check_rotation() returns OK if it is possible to rotate the current
 * tetron clockwise, otherwise it returns BLOCKED.
 */

int check_rotation()
{
   char *c, *limit;
   struct tetron *t;

   t = &tetrons[ current ];
   limit = &t->coordinates[ ( ! t->orientation ? 4 : t->orientation ) * 4 ];

   for( c = &t->coordinates[ ( ! t->orientation ? 3 : t->orientation - 1 ) * 4 ];
        c < limit;
        ++c )
      if ( grid[ row + Y( *c ) ][ column + X( *c ) ] != NULL ||
           ( row + Y( *c )) > 23 || ( column + X( *c )) < 0 ||
           ( column + X( *c )) > 9 )
         return BLOCKED;

   return OK;
}

/*
 * Rotates the current tetron clockwise.
 */

void rotate_tetron()
{
   draw_tetron( OFF );

   if ( --tetrons[ current ].orientation < 0 )
      tetrons[ current ].orientation = 3;

   draw_tetron( ON );
   XFlush( display );
}

/*
 * Exchanges the current tetron for the held tetron.  If there is no
 * held tetron, it holds the current and generates a new one.
 */

void hold_tetron()
{
   char tmp1, tmp2;

   draw_tetron( OFF );

   tmp1 = current;
   tmp2 = tetrons[ current ].orientation;

   if ( pending[ 3 ].type > -1 )
   {
      draw_held( OFF );

      current = pending[ 3 ].type;
      tetrons[ current ].orientation = pending[ 3 ].orientation;

      x = block_size * 3;
      column = 3;

      row = 0;
      y = -( 4 * block_size );
   }
   else
      reset_tetron();

   pending[ 3 ].type = tmp1;
   pending[ 3 ].orientation = tmp2;

   draw_held( ON );
}

/*
 * Recognizes keypresses and dispatches appropriate handler.
 */

void process_key( XKeyEvent *key )
{
   KeySym keysym;

   keysym = XLookupKeysym( key, 0 );

   switch( keysym )
   {
      case XK_Up:
         if ( game_state == PLAY && check_rotation() == OK )
            rotate_tetron();
         break;

      case XK_Down:
         if ( game_state == PLAY )
            drop_tetron();
         break;

      case XK_Left:
         if ( game_state == PLAY && check_motion( LEFT ) == OK )
            move_tetron( LEFT );
         break;

      case XK_Right:
         if ( game_state == PLAY && check_motion( RIGHT ) == OK )
            move_tetron( RIGHT );
         break;

      case XK_p:
         if ( game_state == MENU )
            game_state = PLAY;
         else
            game_state = ( game_state == PAUSE ? PLAY : PAUSE );
         break;

      case XK_q:
         game_state = QUIT;
         break;

      case XK_space:
         if ( game_state == PLAY )
            hold_tetron();
   }
}

/*
 * Gets the next pending event, and reacts accordingly.
 */

void process_event()
{
   Atom protocols, delete;

AGAIN:
   XNextEvent( display, &event );

   if ( XFilterEvent( &event, window ) == True )
      return;

   switch( event.type )
   {
      case Expose:
         if ( event.xexpose.count )
            goto AGAIN;
         else
         {
            draw_game();

            if ( game_state == MENU )
               draw_menu();
            else
            {
               /*
                * Since the drawn contents of the window have disappeared at
                * this point, we need only draw the "on" blocks, resulting in
                * fast screen updating when moving a running game window in a
                * window manager which does "live" moves.
                */

               draw_on();
               if ( tetron_state == OK )
                  draw_tetron( ON );
            }
         }
         break;

      case KeyPress:
         process_key( &event.xkey );
         break;

      case ClientMessage:
         protocols = XInternAtom( display, "WM_PROTOCOLS", False );
         delete = XInternAtom( display, "WM_DELETE_WINDOW", False );

         if ( event.xclient.message_type == protocols &&
              event.xclient.data.l[ 0 ] == delete  &&
              event.xclient.format == 32 )
            game_state = EXIT;
   }
}

/*
 * Moves the current tetron downward one block on the grid.  If the tetron
 * cannot go any farther, it is added to the world grid, and any completed
 * lines of blocks are flashed and removed.
 */

void increment_tetron()
{
   if (( tetron_state = check_motion( DOWN )) == OK )
      move_tetron( DOWN );
   else if ( tetron_state == BLOCKED )
   {
      add_tetron();
      collapse_complete();
   }
}

/*
 * Initializes the pending tetrons array, from the tetrons array,
 * choosing random tetrons and random orientations.
 */

void initialize_pending()
{
   int i;
   float r;

   for( i = 0; i < 3; ++i )
   {
      r = rand();
      pending[ i ].type = r / RAND_MAX * 7;

      r = rand();
      pending[ i ].orientation = r / RAND_MAX * 4;
   }
}

/*
 * Gets a new current tetron from the start of the pending array, and
 * randomly chooses a new tetron for the end of pending tetrons array.
 * Sets the position of the current tetron to the top of the grid.
 */

void reset_tetron()
{
   float r;

   x = block_size * 3;
   column = 3;

   row = 0;
   y = -( 4 * block_size );

   current = pending[ 0 ].type;
   tetrons[ current ].orientation = pending[ 0 ].orientation;

   pending[ 0 ] = pending[ 1 ];
   pending[ 1 ] = pending[ 2 ];

   r = rand();
   pending[ 2 ].type = r / RAND_MAX * 7;

   r = rand();
   pending[ 2 ].orientation = r / RAND_MAX * 4;

   tetron_state = OK;

   draw_game();
   draw_tetron( ON );
   XFlush( display );
}

/*
 * Resets the world grid, the pending tetrons, redraws the screen.
 */

void reset_game()
{
   int i, j;

   score = 0;

   /*
    * The first 4 rows of the world grid are not displayed onscreen.
    * There are twenty rows onscreen.
    */

   for( i = 0; i < 24; ++i )
      for( j = 0; j < 10; ++j )
         grid[ i ][ j ] = NULL;

   initialize_pending();
   pending[ 3 ].type = -1;

   /*
    * Remove the splash screen, and display the new pending
    * tetrons and 0 score.
    */

   draw_window( CLEAR );
   draw_game();
}

/*
 * Displays the GAME OVER message.
 */

void game_over()
{
   char *title = "GAME OVER";
   int w, h, lx, ly, len;

   h = font->ascent + font->descent;
   w = XTextWidth( font, title, len = strlen( title ));
   lx = ( 10 * block_size - w ) / 2;
   ly = ( 20 * block_size - h ) / 2;

   XFillRectangle( display, window, blackGC,
                   small_block_size, ly - block_size,
                   9 * block_size,
                   small_block_size * 3 );

   XDrawRectangle( display, window, whiteGC,
                   small_block_size, ly - block_size,
                   9 * block_size,
                   small_block_size * 3 );

   XDrawString( display, window, whiteGC,
                lx, ly - h / 4, title, len );

   XFlush( display );
   usleep( 999999 );
   usleep( 500001 );
}

/*
 * Main loop.  We process events every 1/40 of a second.  We
 * start out moving the tetrons every 1/2 second, and decrease
 * that interval by 1/40 of a second every 60 seconds.
 */

void play_game()
{
   int ticks, ticker, seconds, fractions;

   ticker = 20;
   fractions = seconds = 0;

   for( ; ; )
   {
      reset_tetron();

      for( ticks = 0; ; )
      {
         if ( XPending( display ) )
            process_event();

         /*
          * If user dropped a tetron (tetron_state == BLOCKED after
          * process_event()), we break out immediately to avoid the delay
          * at usleep()
          */

         if ( tetron_state == BLOCKED )
            break;
         else if( game_state == PAUSE )
            continue;
         else if ( game_state == QUIT || game_state == EXIT )
            return;

         usleep( 25000 );

         if ( ++fractions == 40 )
         {
            ++seconds;
            fractions = 0;
   
            if ( seconds == 60 )
            {
               --ticker;
               seconds = 0;
            }
         }

         if ( ++ticks < ticker )
            continue;

         ticks = 0;
         increment_tetron();

         if ( tetron_state == BLOCKED )
            break;
         else if ( tetron_state == STOPPED )
         {
            game_over();
            return;
         }
      }
   }
}

void game_loop()
{
   for( ; ; )
   {
      game_state = MENU;
      draw_window( CLEAR );
      draw_game();

      while( game_state == MENU )
         process_event();

      if ( game_state == QUIT || game_state == EXIT )
         break;

      reset_game();
      play_game();

      if ( score > highscore )
         highscore = score;

      if ( game_state == EXIT )
         break;
   }
}

int main( int argc, char **argv )
{
   initialize_tetrons();
   initialize_display( argc, argv );

   game_loop();

   cleanup_display();
   return 0;
}
